@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
@@ -16,16 +15,19 @@ from typing import (
 
 from kitty.cli import parse_args
 from kitty.cli_stub import HintsCLIOptions
-from kitty.fast_data_types import set_clipboard_string
-from kitty.key_encoding import (
-    KeyEvent, backspace_key, enter_key, key_defs as K
-)
+from kitty.constants import website_url
+from kitty.fast_data_types import get_options, set_clipboard_string
+from kitty.key_encoding import KeyEvent
 from kitty.typing import BossType, KittyCommonOpts
-from kitty.utils import ScreenSize, screen_size_function, set_primary_selection
+from kitty.utils import (
+    ScreenSize, kitty_ansi_sanitizer_pat, resolve_custom_file,
+    screen_size_function, set_primary_selection
+)
 
 from ..tui.handler import Handler, result_handler
 from ..tui.loop import Loop
 from ..tui.operations import faint, styled
+from ..tui.utils import report_error, report_unhandled_error
 
 
 @lru_cache()
@@ -40,17 +42,27 @@ def kitty_common_opts() -> KittyCommonOpts:
 
 DEFAULT_HINT_ALPHABET = string.digits + string.ascii_lowercase
 DEFAULT_REGEX = r'(?m)^\s*(.+)\s*$'
-ESCAPE = K['ESCAPE']
+FILE_EXTENSION = r'\.(?:[a-zA-Z0-9]{2,7}|[ahcmo])(?!\.)'
+PATH_REGEX = fr'(?:\S*?/[\r\S]+)|(?:\S[\r\S]*{FILE_EXTENSION})\b'
 
 
 class Mark:
 
-    __slots__ = ('index', 'start', 'end', 'text', 'groupdict')
+    __slots__ = ('index', 'start', 'end', 'text', 'is_hyperlink', 'group_id', 'groupdict')
 
-    def __init__(self, index: int, start: int, end: int, text: str, groupdict: Any):
+    def __init__(
+            self,
+            index: int, start: int, end: int,
+            text: str,
+            groupdict: Any,
+            is_hyperlink: bool = False,
+            group_id: Optional[str] = None
+    ):
         self.index, self.start, self.end = index, start, end
         self.text = text
         self.groupdict = groupdict
+        self.is_hyperlink = is_hyperlink
+        self.group_id = group_id
 
 
 @lru_cache(maxsize=2048)
@@ -88,6 +100,11 @@ def highlight_mark(m: Mark, text: str, current_input: str, alphabet: str, colors
     )
 
 
+def debug(*a: Any, **kw: Any) -> None:
+    from ..tui.loop import debug as d
+    d(*a, **kw)
+
+
 def render(text: str, current_input: str, all_marks: Sequence[Mark], ignore_mark_indices: Set[int], alphabet: str, colors: Dict[str, str]) -> str:
     for mark in reversed(all_marks):
         if mark.index in ignore_mark_indices:
@@ -96,12 +113,14 @@ def render(text: str, current_input: str, all_marks: Sequence[Mark], ignore_mark
         text = text[:mark.start] + mtext + text[mark.end:]
 
     text = text.replace('\0', '')
-
-    return text.replace('\n', '\r\n').rstrip()
+    return re.sub('[\r\n]', '\r\n', text).rstrip()
 
 
 class Hints(Handler):
 
+    use_alternate_screen = False  # disabled to avoid screen being blanked at exit causing flicker
+    overlay_ready_report_needed = True
+
     def __init__(self, text: str, all_marks: Sequence[Mark], index_map: Dict[int, Mark], args: HintsCLIOptions):
         self.text, self.index_map = text, index_map
         self.alphabet = args.alphabet or DEFAULT_HINT_ALPHABET
@@ -111,7 +130,7 @@ class Hints(Handler):
         self.all_marks = all_marks
         self.ignore_mark_indices: Set[int] = set()
         self.args = args
-        self.window_title = _('Choose URL') if args.type == 'url' else _('Choose text')
+        self.window_title = args.window_title or (_('Choose URL') if args.type == 'url' else _('Choose text'))
         self.multiple = args.multiple
         self.match_suffix = self.get_match_suffix(args)
         self.chosen: List[Mark] = []
@@ -168,11 +187,11 @@ class Hints(Handler):
             self.draw_screen()
 
     def on_key(self, key_event: KeyEvent) -> None:
-        if key_event is backspace_key:
+        if key_event.matches('backspace'):
             self.current_input = self.current_input[:-1]
             self.current_text = None
             self.draw_screen()
-        elif key_event is enter_key and self.current_input:
+        elif (key_event.matches('enter') or key_event.matches('space')) and self.current_input:
             try:
                 idx = decode_hint(self.current_input, self.alphabet)
                 self.chosen.append(self.index_map[idx])
@@ -187,7 +206,7 @@ class Hints(Handler):
                     self.draw_screen()
                 else:
                     self.quit_loop(0)
-        elif key_event.key is ESCAPE:
+        elif key_event.matches('esc'):
             self.quit_loop(0 if self.multiple else 1)
 
     def on_interrupt(self) -> None:
@@ -206,7 +225,7 @@ class Hints(Handler):
         self.write(self.current_text)
 
 
-def regex_finditer(pat: Pattern, minimum_match_length: int, text: str) -> Generator[Tuple[int, int, Dict], None, None]:
+def regex_finditer(pat: 'Pattern[str]', minimum_match_length: int, text: str) -> Generator[Tuple[int, int, Dict[str, str]], None, None]:
     has_named_groups = bool(pat.groupindex)
     for m in pat.finditer(text):
         s, e = m.span(0 if has_named_groups else pat.groups)
@@ -216,7 +235,7 @@ def regex_finditer(pat: Pattern, minimum_match_length: int, text: str) -> Genera
             yield s, e, m.groupdict()
 
 
-closing_bracket_map = {'(': ')', '[': ']', '{': '}', '<': '>', '*': '*', '"': '"', "'": "'"}
+closing_bracket_map = {'(': ')', '[': ']', '{': '}', '<': '>', '*': '*', '"': '"', "'": "'", "“": "”", "‘": "’"}
 opening_brackets = ''.join(closing_bracket_map)
 PostprocessorFunc = Callable[[str, int, int], Tuple[int, int]]
 postprocessor_map: Dict[str, PostprocessorFunc] = {}
@@ -227,6 +246,11 @@ def postprocessor(func: PostprocessorFunc) -> PostprocessorFunc:
     return func
 
 
+class InvalidMatch(Exception):
+    """Raised when a match turns out to be invalid."""
+    pass
+
+
 @postprocessor
 def url(text: str, s: int, e: int) -> Tuple[int, int]:
     if s > 4 and text[s - 5:s] == 'link:':  # asciidoc URLs
@@ -254,9 +278,13 @@ def brackets(text: str, s: int, e: int) -> Tuple[int, int]:
     # Remove matching brackets
     if s < e <= len(text):
         before = text[s]
-        if before in '({[<' and text[e-1] == closing_bracket_map[before]:
-            s += 1
-            e -= 1
+        if before in '({[<':
+            q = closing_bracket_map[before]
+            if text[e-1] == q:
+                s += 1
+                e -= 1
+            elif text[e:e+1] == q:
+                s += 1
     return s, e
 
 
@@ -265,18 +293,41 @@ def quotes(text: str, s: int, e: int) -> Tuple[int, int]:
     # Remove matching quotes
     if s < e <= len(text):
         before = text[s]
-        if before in '\'"' and text[e-1] == before:
-            s += 1
-            e -= 1
+        if before in '\'"“‘':
+            q = closing_bracket_map[before]
+            if text[e-1] == q:
+                s += 1
+                e -= 1
+            elif text[e:e+1] == q:
+                s += 1
+    return s, e
+
+
+@postprocessor
+def ip(text: str, s: int, e: int) -> Tuple[int, int]:
+    from ipaddress import ip_address
+
+    # Check validity of IPs (or raise InvalidMatch)
+    ip = text[s:e]
+
+    try:
+        ip_address(ip)
+    except Exception:
+        raise InvalidMatch("Invalid IP")
+
     return s, e
 
 
 def mark(pattern: str, post_processors: Iterable[PostprocessorFunc], text: str, args: HintsCLIOptions) -> Generator[Mark, None, None]:
     pat = re.compile(pattern)
     for idx, (s, e, groupdict) in enumerate(regex_finditer(pat, args.minimum_match_length, text)):
-        for func in post_processors:
-            s, e = func(text, s, e)
-        mark_text = text[s:e].replace('\n', '').replace('\0', '')
+        try:
+            for func in post_processors:
+                s, e = func(text, s, e)
+        except InvalidMatch:
+            continue
+
+        mark_text = re.sub('[\r\n\0]', '', text[s:e])
         yield Mark(idx, s, e, mark_text, groupdict)
 
 
@@ -288,7 +339,9 @@ def run_loop(args: HintsCLIOptions, text: str, all_marks: Sequence[Mark], index_
         return {
             'match': handler.text_matches, 'programs': args.program,
             'multiple_joiner': args.multiple_joiner, 'customize_processing': args.customize_processing,
-            'type': args.type, 'groupdicts': handler.groupdicts, 'extra_cli_args': extra_cli_args, 'linenum_action': args.linenum_action
+            'type': args.type, 'groupdicts': handler.groupdicts, 'extra_cli_args': extra_cli_args,
+            'linenum_action': args.linenum_action,
+            'cwd': os.getcwd(),
         }
     raise SystemExit(loop.return_code)
 
@@ -310,17 +363,26 @@ def functions_for(args: HintsCLIOptions) -> Tuple[str, List[PostprocessorFunc]]:
         )
         post_processors.append(url)
     elif args.type == 'path':
-        pattern = r'(?:\S*/\S+)|(?:\S+[.][a-zA-Z0-9]{2,7})'
+        pattern = PATH_REGEX
         post_processors.extend((brackets, quotes))
     elif args.type == 'line':
         pattern = '(?m)^\\s*(.+)[\\s\0]*$'
     elif args.type == 'hash':
-        pattern = '[0-9a-f]{7,128}'
+        pattern = '[0-9a-f][0-9a-f\r]{6,127}'
+    elif args.type == 'ip':
+        pattern = (
+            # # IPv4 with no validation
+            r"((?:\d{1,3}\.){3}\d{1,3}"
+            r"|"
+            # # IPv6 with no validation
+            r"(?:[a-fA-F0-9]{0,4}:){2,7}[a-fA-F0-9]{1,4})"
+        )
+        post_processors.append(ip)
     elif args.type == 'word':
         chars = args.word_characters
         if chars is None:
             chars = kitty_common_opts()['select_by_word_characters']
-        pattern = r'(?u)[{}\w]{{{},}}'.format(escape(chars), args.minimum_match_length)
+        pattern = fr'(?u)[{escape(chars)}\w]{{{args.minimum_match_length},}}'
         post_processors.extend((brackets, quotes))
     else:
         pattern = args.regex
@@ -329,16 +391,22 @@ def functions_for(args: HintsCLIOptions) -> Tuple[str, List[PostprocessorFunc]]:
 
 def convert_text(text: str, cols: int) -> str:
     lines: List[str] = []
-    empty_line = '\0' * cols
+    empty_line = '\0' * cols + '\n'
     for full_line in text.split('\n'):
         if full_line:
             if not full_line.rstrip('\r'):  # empty lines
                 lines.extend(repeat(empty_line, len(full_line)))
                 continue
+            appended = False
             for line in full_line.split('\r'):
                 if line:
                     lines.append(line.ljust(cols, '\0'))
-    return '\n'.join(lines)
+                    lines.append('\r')
+                    appended = True
+            if appended:
+                lines[-1] = '\n'
+    rstripped = re.sub('[\r\n]+$', '', ''.join(lines))
+    return rstripped
 
 
 def parse_input(text: str) -> str:
@@ -352,7 +420,7 @@ def parse_input(text: str) -> str:
 def linenum_marks(text: str, args: HintsCLIOptions, Mark: Type[Mark], extra_cli_args: Sequence[str], *a: Any) -> Generator[Mark, None, None]:
     regex = args.regex
     if regex == DEFAULT_REGEX:
-        regex = r'(?P<path>(?:\S*/\S+?)|(?:\S+[.][a-zA-Z0-9]{2,7})):(?P<line>\d+)'
+        regex = fr'(?P<path>{PATH_REGEX}):(?P<line>\d+)'
     yield from mark(regex, [brackets, quotes], text, args)
 
 
@@ -363,19 +431,70 @@ def load_custom_processor(customize_processing: str) -> Any:
         return {k: getattr(m, k) for k in dir(m)}
     if customize_processing == '::linenum::':
         return {'mark': linenum_marks, 'handle_result': linenum_handle_result}
-    from kitty.constants import resolve_custom_file
     custom_path = resolve_custom_file(customize_processing)
     import runpy
     return runpy.run_path(custom_path, run_name='__main__')
 
 
+def process_escape_codes(text: str) -> Tuple[str, Tuple[Mark, ...]]:
+    hyperlinks: List[Mark] = []
+    removed_size = idx = 0
+    active_hyperlink_url: Optional[str] = None
+    active_hyperlink_id: Optional[str] = None
+    active_hyperlink_start_offset = 0
+
+    def add_hyperlink(end: int) -> None:
+        nonlocal idx, active_hyperlink_url, active_hyperlink_id, active_hyperlink_start_offset
+        assert active_hyperlink_url is not None
+        hyperlinks.append(Mark(
+            idx, active_hyperlink_start_offset, end,
+            active_hyperlink_url,
+            groupdict={},
+            is_hyperlink=True, group_id=active_hyperlink_id
+        ))
+        active_hyperlink_url = active_hyperlink_id = None
+        active_hyperlink_start_offset = 0
+        idx += 1
+
+    def process_hyperlink(m: 're.Match[str]') -> str:
+        nonlocal removed_size, active_hyperlink_url, active_hyperlink_id, active_hyperlink_start_offset
+        raw = m.group()
+        if not raw.startswith('\x1b]8'):
+            removed_size += len(raw)
+            return ''
+        start = m.start() - removed_size
+        removed_size += len(raw)
+        if active_hyperlink_url is not None:
+            add_hyperlink(start)
+        raw = raw[4:-2]
+        parts = raw.split(';', 1)
+        if len(parts) == 2 and parts[1]:
+            active_hyperlink_url = parts[1]
+            active_hyperlink_start_offset = start
+            if parts[0]:
+                for entry in parts[0].split(':'):
+                    if entry.startswith('id=') and len(entry) > 3:
+                        active_hyperlink_id = entry[3:]
+                        break
+
+        return ''
+
+    text = kitty_ansi_sanitizer_pat().sub(process_hyperlink, text)
+    if active_hyperlink_url is not None:
+        add_hyperlink(len(text))
+    return text, tuple(hyperlinks)
+
+
 def run(args: HintsCLIOptions, text: str, extra_cli_args: Sequence[str] = ()) -> Optional[Dict[str, Any]]:
     try:
         text = parse_input(text)
+        text, hyperlinks = process_escape_codes(text)
         pattern, post_processors = functions_for(args)
         if args.type == 'linenum':
             args.customize_processing = '::linenum::'
-        if args.customize_processing:
+        if args.type == 'hyperlink':
+            all_marks = hyperlinks
+        elif args.customize_processing:
             m = load_custom_processor(args.customize_processing)
             if 'mark' in m:
                 all_marks = tuple(m['mark'](text, args, Mark, extra_cli_args))
@@ -384,9 +503,8 @@ def run(args: HintsCLIOptions, text: str, extra_cli_args: Sequence[str] = ()) ->
         else:
             all_marks = tuple(mark(pattern, post_processors, text, args))
         if not all_marks:
-            input(_('No {} found, press Enter to quit.').format(
-                'URLs' if args.type == 'url' else 'matches'
-                ))
+            none_of = {'url': 'URLs', 'hyperlink': 'hyperlinks'}.get(args.type, 'matches')
+            report_error(_('No {} found.').format(none_of))
             return None
 
         largest_index = all_marks[-1].index
@@ -398,11 +516,7 @@ def run(args: HintsCLIOptions, text: str, extra_cli_args: Sequence[str] = ()) ->
                 m.index = largest_index - m.index + offset
         index_map = {m.index: m for m in all_marks}
     except Exception:
-        import traceback
-        traceback.print_exc()
-        input('Press Enter to quit.')
-        raise SystemExit(1)
-
+        report_unhandled_error()
     return run_loop(args, text, all_marks, index_map, extra_cli_args)
 
 
@@ -421,12 +535,12 @@ programs.
 
 --type
 default=url
-choices=url,regex,path,line,hash,word,linenum
+choices=url,regex,path,line,hash,word,linenum,hyperlink,ip
 The type of text to search for. A value of :code:`linenum` is special, it looks
 for error messages using the pattern specified with :option:`--regex`, which
 must have the named groups, :code:`path` and :code:`line`. If not specified,
 will look for :code:`path:line`. The :option:`--linenum-action` option
-controls what to do with the selected error message, other options are ignored.
+controls where to display the selected error message, other options are ignored.
 
 
 --regex
@@ -446,13 +560,15 @@ the form key=value.
 default=self
 type=choice
 choices=self,window,tab,os_window,background
-The action to perform on the matched errors. The actual action is whatever
-arguments are provided to the kitten, for example:
-:code:`kitty + kitten hints --type=linenum vim +{line} {path}`
-will open the matched path at the matched line number in vim. This option
-controls where the action is executed: :code:`self` means the current window,
-:code:`window` a new kitty window, :code:`tab` a new tab, :code:`os_window`
-a new OS window and :code:`background` run in the background.
+Where to perform the action on matched errors. :code:`self` means the current
+window, :code:`window` a new kitty window, :code:`tab` a new tab,
+:code:`os_window` a new OS window and :code:`background` run in the background.
+The actual action is whatever arguments are provided to the kitten, for
+example:
+:code:`kitty + kitten hints --type=linenum --linenum-action=tab vim +{line} {path}`
+will open the matched path at the matched line number in vim in
+a new kitty tab. Note that only when using :code:`self` are the special values for
+:option:`kitty +kitten hints --program` to copy/paste the text respected.
 
 
 --url-prefixes
@@ -494,7 +610,7 @@ indices into the list of selections. You can use 0 for the first selection and
 default=auto
 choices=auto,always,never
 Add trailing space after matched text. Defaults to auto, which adds the space
-when used together with --multiple.
+when used together with :option:`--multiple`.
 
 
 --hints-offset
@@ -537,13 +653,17 @@ The foreground color for text pointed to by the hints
 --customize-processing
 Name of a python file in the kitty config directory which will be imported to provide
 custom implementations for pattern finding and performing actions
-on selected matches. See https://sw.kovidgoyal.net/kitty/kittens/hints.html
+on selected matches. See {hints_url}
 for details. You can also specify absolute paths to load the script from elsewhere.
 
 
+--window-title
+The window title for the hints window, default title is selected based on
+the type of text being hinted.
 '''.format(
     default_regex=DEFAULT_REGEX,
-    line='{{line}}', path='{{path}}'
+    line='{{line}}', path='{{path}}',
+    hints_url=website_url('kittens/hints'),
 ).format
 help_text = 'Select text from the screen using the keyboard. Defaults to searching for URLs.'
 usage = ''
@@ -557,9 +677,7 @@ def main(args: List[str]) -> Optional[Dict[str, Any]]:
     text = ''
     if sys.stdin.isatty():
         if '--help' not in args and '-h' not in args:
-            print('You must pass the text to be hinted on STDIN', file=sys.stderr)
-            input(_('Press Enter to quit'))
-            return None
+            report_unhandled_error('You must pass the text to be hinted on STDIN')
     else:
         text = sys.stdin.buffer.read().decode('utf-8')
         sys.stdin = open(os.ctermid())
@@ -567,23 +685,37 @@ def main(args: List[str]) -> Optional[Dict[str, Any]]:
         opts, items = parse_hints_args(args[1:])
     except SystemExit as e:
         if e.code != 0:
-            print(e.args[0], file=sys.stderr)
-            input(_('Press Enter to quit'))
+            report_unhandled_error(e.args[0])
         return None
     if items and not (opts.customize_processing or opts.type == 'linenum'):
-        print('Extra command line arguments present: {}'.format(' '.join(items)), file=sys.stderr)
-        input(_('Press Enter to quit'))
-    return run(opts, text, items)
+        report_unhandled_error('Extra command line arguments present: {}'.format(' '.join(items)))
+    try:
+        return run(opts, text, items)
+    except Exception:
+        report_unhandled_error()
+    return None
 
 
-def linenum_handle_result(args: List[str], data: Dict[str, Any], target_window_id: int, boss: BossType, extra_cli_args: Sequence[str], *a: Any) -> None:
+def linenum_process_result(data: Dict[str, Any]) -> Tuple[str, int]:
+    pat = re.compile(r':(\d+)$')
     for m, g in zip(data['match'], data['groupdicts']):
         if m:
             path, line = g['path'], g['line']
-            path = os.path.expanduser(path.split(':')[-1])
-            line = int(line)
-            break
-    else:
+            # look for trailers on path of the for :number
+            while True:
+                m = pat.search(path)
+                if m is None:
+                    break
+                line = m.group(1)
+                path = path[:-len(m.group())]
+
+            return os.path.expanduser(path), int(line)
+    return '', -1
+
+
+def linenum_handle_result(args: List[str], data: Dict[str, Any], target_window_id: int, boss: BossType, extra_cli_args: Sequence[str], *a: Any) -> None:
+    path, line = linenum_process_result(data)
+    if not path:
         return
 
     cmd = [x.format(path=path, line=line) for x in extra_cli_args or ('vim', '+{line}', '{path}')]
@@ -592,19 +724,29 @@ def linenum_handle_result(args: List[str], data: Dict[str, Any], target_window_i
 
     if action == 'self':
         if w is not None:
-            import shlex
-            text = ' '.join(shlex.quote(arg) for arg in cmd)
-            w.paste_bytes(text + '\r')
+            is_copy_action = cmd[0] in ('-', '@', '*')
+            if is_copy_action:
+                text = ' '.join(cmd[1:])
+                if cmd[0] == '-':
+                    w.paste_bytes(text)
+                elif cmd[0] == '@':
+                    set_clipboard_string(text)
+                elif cmd[0] == '*':
+                    set_primary_selection(text)
+            else:
+                import shlex
+                text = ' '.join(shlex.quote(arg) for arg in cmd)
+                w.paste_bytes(f'{text}\r')
     elif action == 'background':
         import subprocess
-        subprocess.Popen(cmd)
+        subprocess.Popen(cmd, cwd=data['cwd'])
     else:
         getattr(boss, {
             'window': 'new_window_with_cwd', 'tab': 'new_tab_with_cwd', 'os_window': 'new_os_window_with_cwd'
             }[action])(*cmd)
 
 
-@result_handler(type_of_input='screen')
+@result_handler(type_of_input='screen-ansi', has_ready_notification=Hints.overlay_ready_report_needed)
 def handle_result(args: List[str], data: Dict[str, Any], target_window_id: int, boss: BossType) -> None:
     if data['customize_processing']:
         m = load_custom_processor(data['customize_processing'])
@@ -646,23 +788,26 @@ def handle_result(args: List[str], data: Dict[str, Any], target_window_id: int,
         if program == '-':
             w = boss.window_id_map.get(target_window_id)
             if w is not None:
-                w.paste(joined_text())
+                w.paste_text(joined_text())
         elif program == '@':
             set_clipboard_string(joined_text())
         elif program == '*':
             set_primary_selection(joined_text())
         else:
-            cwd = None
-            w = boss.window_id_map.get(target_window_id)
-            if w is not None:
-                cwd = w.cwd_of_child
-            program = None if program == 'default' else program
-            for m, groupdict in zip(matches, groupdicts):
-                if groupdict:
-                    m = []
-                    for k, v in groupdict.items():
-                        m.append('{}={}'.format(k, v or ''))
-                boss.open_url(m, program, cwd=cwd)
+            cwd = data['cwd']
+            program = get_options().open_url_with if program == 'default' else program
+            if text_type == 'hyperlink':
+                w = boss.window_id_map.get(target_window_id)
+                for m in matches:
+                    if w is not None:
+                        w.open_url(m, hyperlink_id=1, cwd=cwd)
+            else:
+                for m, groupdict in zip(matches, groupdicts):
+                    if groupdict:
+                        m = []
+                        for k, v in groupdict.items():
+                            m.append('{}={}'.format(k, v or ''))
+                    boss.open_url(m, program, cwd=cwd)
 
 
 if __name__ == '__main__':
