@@ -1,14 +1,30 @@
-#!/usr/bin/env python
-# vim:fileencoding=utf-8
+#!/usr/bin/env python3
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
-import os
 import re
 import sys
 from collections import deque
+from typing import (
+    Any, Callable, Dict, FrozenSet, Iterator, List, Match, Optional, Sequence,
+    Tuple, Type, TypeVar, Union, cast
+)
 
+from .cli_stub import CLIOptions
 from .conf.utils import resolve_config
-from .constants import appname, defconf, is_macos, is_wayland, str_version
+from .constants import appname, defconf, is_macos, str_version, website_url
+from .options.types import Options as KittyOpts
+from .typing import BadLineType, TypedDict
+
+
+class OptionDict(TypedDict):
+    dest: str
+    aliases: FrozenSet[str]
+    help: str
+    choices: FrozenSet[str]
+    type: str
+    default: Optional[str]
+    condition: bool
+
 
 CONFIG_HELP = '''\
 Specify a path to the configuration file(s) to use. All configuration files are
@@ -22,95 +38,135 @@ If this option is not specified, config files are searched for in the order:
 :file:`$XDG_CONFIG_DIRS/{appname}/{conf_name}.conf`. The first one that exists is used as the
 config file.
 
-If the environment variable :env:`KITTY_CONFIG_DIRECTORY` is specified, that
+If the environment variable :envvar:`KITTY_CONFIG_DIRECTORY` is specified, that
 directory is always used and the above searching does not happen.
 
 If :file:`/etc/xdg/{appname}/{conf_name}.conf` exists it is merged before (i.e. with lower
 priority) than any user config files. It can be used to specify system-wide
-defaults for all users.
+defaults for all users. You can use either :code:`-` or :code:`/dev/stdin` to read the
+config from STDIN.
 '''.replace(
     '{macos_confpath}',
     (':file:`~/Library/Preferences/{appname}/{conf_name}.conf`,' if is_macos else ''), 1
 )
 
 
-def surround(x, start, end):
+def surround(x: str, start: int, end: int) -> str:
     if sys.stdout.isatty():
-        x = '\033[{}m{}\033[{}m'.format(start, x, end)
+        x = f'\033[{start}m{x}\033[{end}m'
     return x
 
 
-def emph(x):
+role_map: Dict[str, Callable[[str], str]] = {}
+
+
+def role(func: Callable[[str], str]) -> Callable[[str], str]:
+    role_map[func.__name__] = func
+    return func
+
+
+@role
+def emph(x: str) -> str:
     return surround(x, 91, 39)
 
 
-def cyan(x):
+@role
+def cyan(x: str) -> str:
     return surround(x, 96, 39)
 
 
-def green(x):
+@role
+def green(x: str) -> str:
     return surround(x, 32, 39)
 
 
-def blue(x):
+@role
+def blue(x: str) -> str:
     return surround(x, 34, 39)
 
 
-def yellow(x):
+@role
+def yellow(x: str) -> str:
     return surround(x, 93, 39)
 
 
-def italic(x):
+@role
+def italic(x: str) -> str:
     return surround(x, 3, 23)
 
 
-def bold(x):
+@role
+def bold(x: str) -> str:
     return surround(x, 1, 22)
 
 
-def title(x):
+@role
+def title(x: str) -> str:
     return blue(bold(x))
 
 
-def opt(text):
+@role
+def opt(text: str) -> str:
     return text
 
 
-def option(x):
-    idx = x.find('-')
+@role
+def option(x: str) -> str:
+    idx = x.rfind('--')
+    if idx < 0:
+        idx = x.find('-')
     if idx > -1:
         x = x[idx:]
     parts = map(bold, x.split())
     return ' '.join(parts)
 
 
-def code(x):
+@role
+def code(x: str) -> str:
+    return cyan(x)
+
+
+@role
+def kbd(x: str) -> str:
     return x
 
 
-def kbd(x):
-    return x
-
-
-def env(x):
+@role
+def env(x: str) -> str:
     return italic(x)
 
 
-def file(x):
+role_map['envvar'] = role_map['env']
+
+
+@role
+def file(x: str) -> str:
     return italic(x)
 
 
-def parse_option_spec(spec=None):
+@role
+def doc(x: str) -> str:
+    return website_url(x)
+
+
+OptionSpecSeq = List[Union[str, OptionDict]]
+
+
+def parse_option_spec(spec: Optional[str] = None) -> Tuple[OptionSpecSeq, OptionSpecSeq]:
     if spec is None:
         spec = options_spec()
     NORMAL, METADATA, HELP = 'NORMAL', 'METADATA', 'HELP'
     state = NORMAL
     lines = spec.splitlines()
     prev_line = ''
-    seq = []
-    disabled = []
+    seq: OptionSpecSeq = []
+    disabled: OptionSpecSeq = []
     mpat = re.compile('([a-z]+)=(.+)')
-    current_cmd = None
+    current_cmd: OptionDict = {
+        'dest': '', 'aliases': frozenset(), 'help': '', 'choices': frozenset(),
+        'type': '', 'condition': False, 'default': None
+    }
+    empty_cmd = current_cmd
 
     for line in lines:
         line = line.rstrip()
@@ -122,10 +178,14 @@ def parse_option_spec(spec=None):
                 continue
             if line.startswith('--'):
                 parts = line.split(' ')
-                current_cmd = {'dest': parts[0][2:].replace('-', '_'), 'aliases': frozenset(parts), 'help': ''}
+                current_cmd = {
+                    'dest': parts[0][2:].replace('-', '_'), 'aliases': frozenset(parts), 'help': '',
+                    'choices': frozenset(), 'type': '',
+                    'default': None, 'condition': True
+                }
                 state = METADATA
                 continue
-            raise ValueError('Invalid option spec, unexpected line: {}'.format(line))
+            raise ValueError(f'Invalid option spec, unexpected line: {line}')
         elif state is METADATA:
             m = mpat.match(line)
             if m is None:
@@ -133,11 +193,17 @@ def parse_option_spec(spec=None):
                 current_cmd['help'] += line
             else:
                 k, v = m.group(1), m.group(2)
-                if k == 'condition':
-                    v = eval(v)
-                current_cmd[k] = v
                 if k == 'choices':
-                    current_cmd['choices'] = {x.strip() for x in current_cmd['choices'].split(',')}
+                    current_cmd['choices'] = frozenset(x.strip() for x in v.split(','))
+                else:
+                    if k == 'default':
+                        current_cmd['default'] = v
+                    elif k == 'type':
+                        current_cmd['type'] = v
+                    elif k == 'dest':
+                        current_cmd['dest'] = v
+                    elif k == 'condition':
+                        current_cmd['condition'] = bool(eval(v))
         elif state is HELP:
             if line:
                 spc = '' if current_cmd['help'].endswith('\n') else ' '
@@ -148,38 +214,40 @@ def parse_option_spec(spec=None):
                 else:
                     state = NORMAL
                     (seq if current_cmd.get('condition', True) else disabled).append(current_cmd)
-                    current_cmd = None
+                    current_cmd = empty_cmd
         prev_line = line
-    if current_cmd is not None:
+    if current_cmd is not empty_cmd:
         (seq if current_cmd.get('condition', True) else disabled).append(current_cmd)
 
     return seq, disabled
 
 
-def prettify(text):
-    role_map = globals()
+def prettify(text: str) -> str:
 
-    def sub(m):
+    def identity(x: str) -> str:
+        return x
+
+    def sub(m: 'Match[str]') -> str:
         role, text = m.group(1, 2)
-        return role_map[role](text)
+        return role_map.get(role, identity)(text)
 
     text = re.sub(r':([a-z]+):`([^`]+)`', sub, text)
     return text
 
 
-def prettify_rst(text):
+def prettify_rst(text: str) -> str:
     return re.sub(r':([a-z]+):`([^`]+)`(=[^\s.]+)', r':\1:`\2`:code:`\3`', text)
 
 
-def version(add_rev=False):
+def version(add_rev: bool = False) -> str:
     rev = ''
     from . import fast_data_types
     if add_rev and hasattr(fast_data_types, 'KITTY_VCS_REV'):
-        rev = ' ({})'.format(fast_data_types.KITTY_VCS_REV[:10])
+        rev = f' ({fast_data_types.KITTY_VCS_REV[:10]})'
     return '{} {}{} created by {}'.format(italic(appname), green(str_version), rev, title('Kovid Goyal'))
 
 
-def wrap(text, limit=80):
+def wrap(text: str, limit: int = 80) -> Iterator[str]:
     NORMAL, IN_FORMAT = 'NORMAL', 'IN_FORMAT'
     state = NORMAL
     last_space_at = None
@@ -203,7 +271,7 @@ def wrap(text, limit=80):
             last_space_at = None
             chars_in_line = i - breaks[-1]
 
-    lines = []
+    lines: List[str] = []
     for b in reversed(breaks):
         lines.append(text[b:].lstrip())
         text = text[:b]
@@ -212,81 +280,103 @@ def wrap(text, limit=80):
     return reversed(lines)
 
 
+def get_defaults_from_seq(seq: OptionSpecSeq) -> Dict[str, Any]:
+    ans: Dict[str, Any] = {}
+    for opt in seq:
+        if not isinstance(opt, str):
+            ans[opt['dest']] = defval_for_opt(opt)
+    return ans
+
+
 default_msg = ('''\
 Run the :italic:`{appname}` terminal emulator. You can also specify the :italic:`program`
 to run inside :italic:`{appname}` as normal arguments following the :italic:`options`.
-For example: {appname} /bin/sh
+For example: {appname} sh -c "echo hello, world. Press ENTER to quit; read"
 
-For comprehensive documentation for kitty, please see: https://sw.kovidgoyal.net/kitty''').format(appname=appname)
+For comprehensive documentation for kitty, please see: {url}''').format(
+    appname=appname, url=website_url())
 
 
-def print_help_for_seq(seq, usage, message, appname):
-    from kitty.utils import screen_size_function
-    screen_size = screen_size_function()
-    try:
-        linesz = min(screen_size().cols, 76)
-    except EnvironmentError:
-        linesz = 76
-    blocks = []
-    a = blocks.append
+class PrintHelpForSeq:
 
-    def wa(text, indent=0, leading_indent=None):
-        if leading_indent is None:
-            leading_indent = indent
-        j = '\n' + (' ' * indent)
-        lines = []
-        for l in text.splitlines():
-            if l:
-                lines.extend(wrap(l, limit=linesz - indent))
-            else:
-                lines.append('')
-        a((' ' * leading_indent) + j.join(lines))
+    allow_pager = True
 
-    usage = '[program-to-run ...]' if usage is None else usage
-    optstring = '[options] ' if seq else ''
-    a('{}: {} {}{}'.format(title('Usage'), bold(yellow(appname)), optstring, usage))
-    a('')
-    message = message or default_msg
-    wa(prettify(message))
-    a('')
-    if seq:
-        a('{}:'.format(title('Options')))
-    for opt in seq:
-        if isinstance(opt, str):
-            a('{}:'.format(title(opt)))
-            continue
-        help_text = opt['help']
-        if help_text == '!':
-            continue  # hidden option
-        a('  ' + ', '.join(map(green, sorted(opt['aliases']))))
-        if not opt.get('type', '').startswith('bool-'):
-            blocks[-1] += '={}'.format(italic(opt['dest'].upper()))
-        if opt.get('help'):
-            defval = opt.get('default')
-            t = help_text.replace('%default', str(defval))
-            wa(prettify(t.strip()), indent=4)
-            if defval is not None:
-                wa('Default: {}'.format(defval), indent=4)
-            if 'choices' in opt:
-                wa('Choices: {}'.format(', '.join(opt['choices'])), indent=4)
-            a('')
+    def __call__(self, seq: OptionSpecSeq, usage: Optional[str], message: Optional[str], appname: str) -> None:
+        from kitty.utils import screen_size_function
+        screen_size = screen_size_function()
+        try:
+            linesz = min(screen_size().cols, 76)
+        except OSError:
+            linesz = 76
+        blocks: List[str] = []
+        a = blocks.append
 
-    text = '\n'.join(blocks) + '\n\n' + version()
-    if print_help_for_seq.allow_pager and sys.stdout.isatty():
-        import subprocess
-        p = subprocess.Popen(['less', '-isRXF'], stdin=subprocess.PIPE)
-        p.communicate(text.encode('utf-8'))
-        raise SystemExit(p.wait())
-    else:
-        print(text)
+        def wa(text: str, indent: int = 0, leading_indent: Optional[int] = None) -> None:
+            if leading_indent is None:
+                leading_indent = indent
+            j = '\n' + (' ' * indent)
+            lines: List[str] = []
+            for ln in text.splitlines():
+                if ln:
+                    lines.extend(wrap(ln, limit=linesz - indent))
+                else:
+                    lines.append('')
+            a((' ' * leading_indent) + j.join(lines))
+
+        usage = '[program-to-run ...]' if usage is None else usage
+        optstring = '[options] ' if seq else ''
+        a('{}: {} {}{}'.format(title('Usage'), bold(yellow(appname)), optstring, usage))
+        a('')
+        message = message or default_msg
+        wa(prettify(message))
+        a('')
+        if seq:
+            a('{}:'.format(title('Options')))
+        for opt in seq:
+            if isinstance(opt, str):
+                a(f'{title(opt)}:')
+                continue
+            help_text = opt['help']
+            if help_text == '!':
+                continue  # hidden option
+            a('  ' + ', '.join(map(green, sorted(opt['aliases']))))
+            if not opt.get('type', '').startswith('bool-'):
+                blocks[-1] += '={}'.format(italic(opt['dest'].upper()))
+            if opt.get('help'):
+                defval = opt.get('default')
+                t = help_text.replace('%default', str(defval))
+                wa(prettify(t.strip()), indent=4)
+                if defval is not None:
+                    wa(f'Default: {defval}', indent=4)
+                if opt.get('choices'):
+                    wa('Choices: {}'.format(', '.join(opt['choices'])), indent=4)
+                a('')
+
+        text = '\n'.join(blocks) + '\n\n' + version()
+        if print_help_for_seq.allow_pager and sys.stdout.isatty():
+            import subprocess
+            p = subprocess.Popen(['less', '-isRXF'], stdin=subprocess.PIPE)
+            try:
+                p.communicate(text.encode('utf-8'))
+            except KeyboardInterrupt:
+                raise SystemExit(1)
+            raise SystemExit(p.wait())
+        else:
+            print(text)
 
 
-print_help_for_seq.allow_pager = True
+print_help_for_seq = PrintHelpForSeq()
 
 
-def seq_as_rst(seq, usage, message, appname, heading_char='-'):
+def seq_as_rst(
+    seq: OptionSpecSeq,
+    usage: Optional[str],
+    message: Optional[str],
+    appname: Optional[str],
+    heading_char: str = '-'
+) -> str:
     import textwrap
-    blocks = []
+    blocks: List[str] = []
     a = blocks.append
 
     usage = '[program-to-run ...]' if usage is None else usage
@@ -294,7 +384,7 @@ def seq_as_rst(seq, usage, message, appname, heading_char='-'):
     a('.. highlight:: sh')
     a('.. code-block:: sh')
     a('')
-    a('  {} {}{}'.format(appname, optstring, usage))
+    a(f'  {appname} {optstring}{usage}')
     a('')
     message = message or default_msg
     a(prettify_rst(message))
@@ -322,17 +412,46 @@ def seq_as_rst(seq, usage, message, appname, heading_char='-'):
             a('')
             a(textwrap.indent(prettify_rst(t), ' ' * 4))
             if defval is not None:
-                a(textwrap.indent('Default: :code:`{}`'.format(defval), ' ' * 4))
-            if 'choices' in opt:
-                a(textwrap.indent('Choices: :code:`{}`'.format(', '.join(sorted(opt['choices']))), ' ' * 4))
+                a(textwrap.indent(f'Default: :code:`{defval}`', ' ' * 4))
+            if opt.get('choices'):
+                a(textwrap.indent('Choices: {}'.format(', '.join(f':code:`{c}`' for c in sorted(opt['choices']))), ' ' * 4))
             a('')
 
     text = '\n'.join(blocks)
     return text
 
 
-def defval_for_opt(opt):
-    dv = opt.get('default')
+def as_type_stub(seq: OptionSpecSeq, disabled: OptionSpecSeq, class_name: str, extra_fields: Sequence[str] = ()) -> str:
+    from itertools import chain
+    ans: List[str] = [f'class {class_name}:']
+    for opt in chain(seq, disabled):
+        if isinstance(opt, str):
+            continue
+        name = opt['dest']
+        otype = opt['type'] or 'str'
+        if otype in ('str', 'int', 'float'):
+            t = otype
+            if t == 'str' and defval_for_opt(opt) is None:
+                t = 'typing.Optional[str]'
+        elif otype == 'list':
+            t = 'typing.Sequence[str]'
+        elif otype in ('choice', 'choices'):
+            if opt['choices']:
+                t = 'typing.Literal[{}]'.format(','.join(f'{x!r}' for x in opt['choices']))
+            else:
+                t = 'str'
+        elif otype.startswith('bool-'):
+            t = 'bool'
+        else:
+            raise ValueError(f'Unknown CLI option type: {otype}')
+        ans.append(f'    {name}: {t}')
+    for x in extra_fields:
+        ans.append(f'    {x}')
+    return '\n'.join(ans) + '\n\n\n'
+
+
+def defval_for_opt(opt: OptionDict) -> Any:
+    dv: Any = opt.get('default')
     typ = opt.get('type', '')
     if typ.startswith('bool-'):
         if dv is None:
@@ -348,11 +467,11 @@ def defval_for_opt(opt):
 
 class Options:
 
-    def __init__(self, seq, usage, message, appname):
+    def __init__(self, seq: OptionSpecSeq, usage: Optional[str], message: Optional[str], appname: Optional[str]):
         self.alias_map = {}
         self.seq = seq
-        self.names_map = {}
-        self.values_map = {}
+        self.names_map: Dict[str, OptionDict] = {}
+        self.values_map: Dict[str, Any] = {}
         self.usage, self.message, self.appname = usage, message, appname
         for opt in seq:
             if isinstance(opt, str):
@@ -363,13 +482,13 @@ class Options:
             self.names_map[name] = opt
             self.values_map[name] = defval_for_opt(opt)
 
-    def opt_for_alias(self, alias):
+    def opt_for_alias(self, alias: str) -> OptionDict:
         opt = self.alias_map.get(alias)
         if opt is None:
-            raise SystemExit('Unknown option: {}'.format(emph(alias)))
+            raise SystemExit(f'Unknown option: {emph(alias)}')
         return opt
 
-    def needs_arg(self, alias):
+    def needs_arg(self, alias: str) -> bool:
         if alias in ('-h', '--help'):
             print_help_for_seq(self.seq, self.usage, self.message, self.appname or appname)
             raise SystemExit(0)
@@ -380,7 +499,7 @@ class Options:
         typ = opt.get('type', '')
         return not typ.startswith('bool-')
 
-    def process_arg(self, alias, val=None):
+    def process_arg(self, alias: str, val: Any = None) -> None:
         opt = self.opt_for_alias(alias)
         typ = opt.get('type', '')
         name = opt['dest']
@@ -409,32 +528,25 @@ class Options:
             self.values_map[name] = val
 
 
-class Namespace:
-
-    def __init__(self, kwargs):
-        for name in kwargs:
-            setattr(self, name, kwargs[name])
-
-
-def parse_cmdline(oc, disabled, args=None):
+def parse_cmdline(oc: Options, disabled: OptionSpecSeq, ans: Any, args: Optional[List[str]] = None) -> List[str]:
     NORMAL, EXPECTING_ARG = 'NORMAL', 'EXPECTING_ARG'
     state = NORMAL
-    if args is None:
-        args = sys.argv[1:]
-    args = deque(args)
+    dargs = deque(sys.argv[1:] if args is None else args)
+    leftover_args: List[str] = []
     current_option = None
 
-    while args:
-        arg = args.popleft()
+    while dargs:
+        arg = dargs.popleft()
         if state is NORMAL:
             if arg.startswith('-'):
                 if arg == '--':
+                    leftover_args = list(dargs)
                     break
                 parts = arg.split('=', 1)
                 needs_arg = oc.needs_arg(parts[0])
                 if not needs_arg:
                     if len(parts) != 1:
-                        raise SystemExit('The {} option does not accept arguments'.format(emph(parts[0])))
+                        raise SystemExit(f'The {emph(parts[0])} option does not accept arguments')
                     oc.process_arg(parts[0])
                     continue
                 if len(parts) == 1:
@@ -443,21 +555,23 @@ def parse_cmdline(oc, disabled, args=None):
                     continue
                 oc.process_arg(parts[0], parts[1])
             else:
-                args = [arg] + list(args)
+                leftover_args = [arg] + list(dargs)
                 break
-        else:
+        elif current_option is not None:
             oc.process_arg(current_option, arg)
             current_option, state = None, NORMAL
     if state is EXPECTING_ARG:
-        raise SystemExit('An argument is required for the option: {}'.format(emph(arg)))
+        raise SystemExit(f'An argument is required for the option: {emph(arg)}')
 
-    ans = Namespace(oc.values_map)
+    for key, val in oc.values_map.items():
+        setattr(ans, key, val)
     for opt in disabled:
-        setattr(ans, opt['dest'], defval_for_opt(opt))
-    return ans, list(args)
+        if not isinstance(opt, str):
+            setattr(ans, opt['dest'], defval_for_opt(opt))
+    return leftover_args
 
 
-def options_spec():
+def options_spec() -> str:
     if not hasattr(options_spec, 'ans'):
         OPTIONS = '''
 --class
@@ -473,10 +587,8 @@ Set the name part of the :italic:`WM_CLASS` property (defaults to using the valu
 
 
 --title -T
-Set the window title. This will override any title set by the program running inside kitty. So
-only use this if you are running a program that does not set titles. If combined
-with :option:`{appname} --session` the title will be used for all windows created by the
-session, that do not set their own titles.
+Set the OS window title. This will override any title set by the program running inside kitty, permanently
+fixing the OS Window's title. So only use this if you are running a program that does not set titles.
 
 
 --config -c
@@ -490,7 +602,7 @@ Override individual configuration options, can be specified multiple times.
 Syntax: :italic:`name=value`. For example: :option:`kitty -o` font_size=20
 
 
---directory -d
+--directory --working-directory -d
 default=.
 Change to the specified directory when launching
 
@@ -503,13 +615,13 @@ Detach from the controlling terminal, if any
 
 --session
 Path to a file containing the startup :italic:`session` (tabs, windows, layout, programs).
-See the README file for details and an example.
+Use - to read from STDIN. See the README file for details and an example.
 
 
 --hold
 type=bool-set
 Remain open after child process exits. Note that this only affects the first
-window. You can quit by either using the close window shortcut or :kbd:`Ctrl+d`.
+window. You can quit by either using the close window shortcut or pressing any key.
 
 
 --single-instance -1
@@ -540,12 +652,15 @@ Tell kitty to listen on the specified address for control
 messages. For example, :option:`{appname} --listen-on`=unix:/tmp/mykitty or
 :option:`{appname} --listen-on`=tcp:localhost:12345. On Linux systems, you can
 also use abstract UNIX sockets, not associated with a file, like this:
-:option:`{appname} --listen-on`=unix:@mykitty.  To control kitty, you can send
+:option:`{appname} --listen-on`=unix:@mykitty. Environment variables
+in the setting are expanded and relative paths are resolved with
+respect to the temporary directory. To control kitty, you can send
 it commands with :italic:`kitty @` using the :option:`kitty @ --to` option to
 specify this address. This option will be ignored, unless you set
 :opt:`allow_remote_control` to yes in :file:`kitty.conf`. Note that if you run
-:italic:`kitty @` within a kitty window, there is no need to specify the :italic:`--to`
-option as it is read automatically from the environment.
+:italic:`kitty @` within a kitty window, there is no need to specify the :option:`kitty @ --to`
+option as it is read automatically from the environment. For UNIX sockets, this
+can also be specified in :file:`kitty.conf`.
 
 
 --start-as
@@ -568,7 +683,7 @@ Output commands received from child process to stdout
 
 
 --replay-commands
-Replay previously dumped commands. Specify the path to a dump file previously created by --dump-commands. You
+Replay previously dumped commands. Specify the path to a dump file previously created by :option:`kitty --dump-commands`. You
 can open a new kitty window to replay the commands with::
 
     kitty sh -c "kitty --replay-commands /path/to/dump/file; read"
@@ -578,15 +693,17 @@ can open a new kitty window to replay the commands with::
 Path to file in which to store the raw bytes received from the child process
 
 
---debug-gl
+--debug-rendering --debug-gl
 type=bool-set
-Debug OpenGL commands. This will cause all OpenGL calls to check for errors
-instead of ignoring them. Useful when debugging rendering problems
+Debug rendering commands. This will cause all OpenGL calls to check for errors
+instead of ignoring them. Also prints out miscellaneous debug information.
+Useful when debugging rendering problems
 
 
---debug-keyboard
+--debug-input --debug-keyboard
+dest=debug_keyboard
 type=bool-set
-This option will cause kitty to print out key events as they are received
+This option will cause kitty to print out key and mouse events as they are received
 
 
 --debug-font-fallback
@@ -594,137 +711,74 @@ type=bool-set
 Print out information about the selection of fallback fonts for characters not present in the main font.
 
 
---debug-config
-type=bool-set
-Print out information about the system and kitty configuration.
+--watcher
+This option is deprecated in favor of the :opt:`watcher` option in kitty.conf and should not be used.
 
 
 --execute -e
 type=bool-set
 !
 '''
-        options_spec.ans = OPTIONS.format(
-            appname=appname, config_help=CONFIG_HELP.format(appname=appname, conf_name=appname)
-
-        )
-    return options_spec.ans
+        setattr(options_spec, 'ans', OPTIONS.format(
+            appname=appname, config_help=CONFIG_HELP.format(appname=appname, conf_name=appname),
+        ))
+    ans: str = getattr(options_spec, 'ans')
+    return ans
 
 
-def options_for_completion():
-    return parse_option_spec(options_spec())[0]
+def options_for_completion() -> OptionSpecSeq:
+    raw = '--help -h\ntype=bool-set\nShow help for {appname} command line options\n\n{raw}'.format(
+            appname=appname, raw=options_spec())
+    return parse_option_spec(raw)[0]
 
 
-def option_spec_as_rst(ospec=options_spec, usage=None, message=None, appname=None, heading_char='-'):
+def option_spec_as_rst(
+    ospec: Callable[[], str] = options_spec,
+    usage: Optional[str] = None, message: Optional[str] = None, appname: Optional[str] = None,
+    heading_char: str = '-'
+) -> str:
     options = parse_option_spec(ospec())
     seq, disabled = options
     oc = Options(seq, usage, message, appname)
     return seq_as_rst(oc.seq, oc.usage, oc.message, oc.appname, heading_char=heading_char)
 
 
-def parse_args(args=None, ospec=options_spec, usage=None, message=None, appname=None):
+T = TypeVar('T')
+
+
+def parse_args(
+    args: Optional[List[str]] = None,
+    ospec: Callable[[], str] = options_spec,
+    usage: Optional[str] = None,
+    message: Optional[str] = None,
+    appname: Optional[str] = None,
+    result_class: Optional[Type[T]] = None,
+) -> Tuple[T, List[str]]:
     options = parse_option_spec(ospec())
     seq, disabled = options
     oc = Options(seq, usage, message, appname)
-    return parse_cmdline(oc, disabled, args=args)
+    if result_class is not None:
+        ans = result_class()
+    else:
+        ans = cast(T, CLIOptions())
+    return ans, parse_cmdline(oc, disabled, ans, args=args)
 
 
 SYSTEM_CONF = '/etc/xdg/kitty/kitty.conf'
 
 
-def print_shortcut(key_sequence, action):
-    if not getattr(print_shortcut, 'maps', None):
-        from kitty.keys import defines
-        v = vars(defines)
-        mmap = {m[len('GLFW_MOD_'):].lower(): x for m, x in v.items() if m.startswith('GLFW_MOD_')}
-        kmap = {k[len('GLFW_KEY_'):].lower(): x for k, x in v.items() if k.startswith('GLFW_KEY_')}
-        krmap = {v: k for k, v in kmap.items()}
-        print_shortcut.maps = mmap, krmap
-    mmap, krmap = print_shortcut.maps
-    keys = []
-    for key in key_sequence:
-        names = []
-        mods, is_native, key = key
-        for name, val in mmap.items():
-            if mods & val:
-                names.append(name)
-        if key:
-            if is_native:
-                from .fast_data_types import GLFW_KEY_UNKNOWN, glfw_get_key_name
-                names.append(glfw_get_key_name(GLFW_KEY_UNKNOWN, key))
-            else:
-                names.append(krmap[key])
-        keys.append('+'.join(names))
-
-    print('\t', ' > '.join(keys), action)
-
-
-def print_shortcut_changes(defns, text, changes):
-    if changes:
-        print(title(text))
-
-        for k in sorted(changes):
-            print_shortcut(k, defns[k])
-
-
-def compare_keymaps(final, initial):
-    added = set(final) - set(initial)
-    removed = set(initial) - set(final)
-    changed = {k for k in set(final) & set(initial) if final[k] != initial[k]}
-    print_shortcut_changes(final, 'Added shortcuts:', added)
-    print_shortcut_changes(initial, 'Removed shortcuts:', removed)
-    print_shortcut_changes(final, 'Changed shortcuts:', changed)
-
-
-def flatten_sequence_map(m):
-    ans = {}
-    for k, rest_map in m.items():
-        for r, action in rest_map.items():
-            ans[(k,) + (r)] = action
-    return ans
-
-
-def compare_opts(opts):
-    from .config import defaults, load_config
-    print('\nConfig options different from defaults:')
-    default_opts = load_config()
-    changed_opts = [
-        f for f in sorted(defaults._fields)
-        if f not in ('key_definitions', 'keymap', 'sequence_map') and getattr(opts, f) != getattr(defaults, f)
-    ]
-    field_len = max(map(len, changed_opts)) if changed_opts else 20
-    fmt = '{{:{:d}s}}'.format(field_len)
-    for f in changed_opts:
-        print(title(fmt.format(f)), getattr(opts, f))
-
-    final, initial = opts.keymap, default_opts.keymap
-    final = {(k,): v for k, v in final.items()}
-    initial = {(k,): v for k, v in initial.items()}
-    final_s, initial_s = map(flatten_sequence_map, (opts.sequence_map, default_opts.sequence_map))
-    final.update(final_s)
-    initial.update(initial_s)
-    compare_keymaps(final, initial)
-
-
-def create_opts(args, debug_config=False):
+def create_opts(args: CLIOptions, accumulate_bad_lines: Optional[List[BadLineType]] = None) -> KittyOpts:
     from .config import load_config
     config = tuple(resolve_config(SYSTEM_CONF, defconf, args.config))
-    if debug_config:
-        print(version(add_rev=True))
-        print(' '.join(os.uname()))
-        if is_macos:
-            import subprocess
-            print(' '.join(subprocess.check_output(['sw_vers']).decode('utf-8').splitlines()).strip())
-        else:
-            print('Running under:', green('Wayland' if is_wayland else 'X11'))
-        if os.path.exists('/etc/issue'):
-            print(open('/etc/issue', encoding='utf-8', errors='replace').read().strip())
-        if os.path.exists('/etc/lsb-release'):
-            print(open('/etc/lsb-release', encoding='utf-8', errors='replace').read().strip())
-        config = tuple(x for x in config if os.path.exists(x))
-        if config:
-            print(green('Loaded config files:'), ', '.join(config))
-    overrides = (a.replace('=', ' ', 1) for a in args.override or ())
-    opts = load_config(*config, overrides=overrides)
-    if debug_config:
-        compare_opts(opts)
+    # Does not cover the case where `name =` when `=` is the value.
+    pat = re.compile(r'^([a-zA-Z0-9_]+)[ \t]*=')
+    overrides = (pat.sub(r'\1 ', a.lstrip()) for a in args.override or ())
+    opts = load_config(*config, overrides=overrides, accumulate_bad_lines=accumulate_bad_lines)
+    return opts
+
+
+def create_default_opts() -> KittyOpts:
+    from .config import load_config
+    config = tuple(resolve_config(SYSTEM_CONF, defconf, ()))
+    opts = load_config(*config)
     return opts
