@@ -1,41 +1,55 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
+import os
+import stat
 import weakref
 from collections import deque
 from contextlib import suppress
-from functools import partial
 from operator import attrgetter
+from time import monotonic
 from typing import (
-    Any, Deque, Dict, Generator, Iterator, List, NamedTuple, Optional, Pattern,
-    Sequence, Tuple, Union, cast
+    Any, Deque, Dict, Generator, Iterable, Iterator, List, NamedTuple,
+    Optional, Pattern, Sequence, Set, Tuple, Union, cast
 )
 
-from .borders import Borders
+from .borders import Border, Borders
 from .child import Child
 from .cli_stub import CLIOptions
-from .constants import appname
+from .constants import appname, kitty_exe
 from .fast_data_types import (
-    add_tab, attach_window, detach_window, get_boss, mark_tab_bar_dirty,
-    next_window_id, remove_tab, remove_window, ring_bell, set_active_tab,
-    set_active_window, swap_tabs, sync_os_window_title
+    GLFW_MOUSE_BUTTON_LEFT, GLFW_MOUSE_BUTTON_MIDDLE, GLFW_PRESS, GLFW_RELEASE,
+    add_tab, attach_window, detach_window, get_boss, get_click_interval,
+    get_options, mark_tab_bar_dirty, next_window_id, remove_tab, remove_window,
+    ring_bell, set_active_tab, set_active_window, swap_tabs,
+    sync_os_window_title
 )
-from .layout.base import Layout, Rect
+from .layout.base import Layout
 from .layout.interface import create_layout_object_for, evict_cached_layouts
-from .options_stub import Options
 from .tab_bar import TabBar, TabBarData
+from .types import ac
 from .typing import EdgeLiteral, SessionTab, SessionType, TypedDict
 from .utils import log_error, platform_window_id, resolved_shell
-from .window import Watchers, Window, WindowDict
+from .window import CwdRequest, Watchers, Window, WindowDict
 from .window_list import WindowList
 
 
+class TabMouseEvent(NamedTuple):
+    button: int
+    modifiers: int
+    action: int
+    at: float
+    tab_idx: Optional[int]
+
+
 class TabDict(TypedDict):
     id: int
     is_focused: bool
     title: str
     layout: str
+    layout_state: Dict[str, Any]
+    layout_opts: Dict[str, Any]
+    enabled_layouts: List[str]
     windows: List[WindowDict]
     active_window_history: List[int]
 
@@ -44,24 +58,26 @@ class SpecialWindowInstance(NamedTuple):
     cmd: Optional[List[str]]
     stdin: Optional[bytes]
     override_title: Optional[str]
-    cwd_from: Optional[int]
+    cwd_from: Optional[CwdRequest]
     cwd: Optional[str]
     overlay_for: Optional[int]
     env: Optional[Dict[str, str]]
     watchers: Optional[Watchers]
+    overlay_behind: bool
 
 
 def SpecialWindow(
     cmd: Optional[List[str]],
     stdin: Optional[bytes] = None,
     override_title: Optional[str] = None,
-    cwd_from: Optional[int] = None,
+    cwd_from: Optional[CwdRequest] = None,
     cwd: Optional[str] = None,
     overlay_for: Optional[int] = None,
     env: Optional[Dict[str, str]] = None,
-    watchers: Optional[Watchers] = None
+    watchers: Optional[Watchers] = None,
+    overlay_behind: bool = False
 ) -> SpecialWindowInstance:
-    return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env, watchers)
+    return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env, watchers, overlay_behind)
 
 
 def add_active_id_to_history(items: Deque[int], item_id: int, maxlen: int = 64) -> None:
@@ -74,26 +90,29 @@ def add_active_id_to_history(items: Deque[int], item_id: int, maxlen: int = 64)
 
 class Tab:  # {{{
 
+    active_fg: Optional[int] = None
+    active_bg: Optional[int] = None
+    inactive_fg: Optional[int] = None
+    inactive_bg: Optional[int] = None
+
     def __init__(
         self,
         tab_manager: 'TabManager',
         session_tab: Optional['SessionTab'] = None,
         special_window: Optional[SpecialWindowInstance] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         no_initial_window: bool = False
     ):
         self.tab_manager_ref = weakref.ref(tab_manager)
         self.os_window_id: int = tab_manager.os_window_id
         self.id: int = add_tab(self.os_window_id)
         if not self.id:
-            raise Exception('No OS window with id {} found, or tab counter has wrapped'.format(self.os_window_id))
-        self.opts, self.args = tab_manager.opts, tab_manager.args
+            raise Exception(f'No OS window with id {self.os_window_id} found, or tab counter has wrapped')
+        self.args = tab_manager.args
         self.name = getattr(session_tab, 'name', '')
-        self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or self.opts.enabled_layouts]
-        self.borders = Borders(self.os_window_id, self.id, self.opts)
+        self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or get_options().enabled_layouts]
+        self.borders = Borders(self.os_window_id, self.id)
         self.windows = WindowList(self)
-        for i, which in enumerate('first second third fourth fifth sixth seventh eighth ninth tenth'.split()):
-            setattr(self, which + '_window', partial(self.nth_window, num=i))
         self._last_used_layout: Optional[str] = None
         self._current_layout_name: Optional[str] = None
         self.cwd = self.args.directory
@@ -113,6 +132,17 @@ class Tab:  # {{{
             self._set_current_layout(l0)
             self.startup(session_tab)
 
+    def set_enabled_layouts(self, val: Iterable[str]) -> None:
+        self.enabled_layouts = [x.lower() for x in val] or ['tall']
+        if self.current_layout.name not in self.enabled_layouts:
+            self._set_current_layout(self.enabled_layouts[0])
+        self.relayout()
+
+    def apply_options(self) -> None:
+        for window in self:
+            window.apply_options()
+        self.set_enabled_layouts(get_options().enabled_layouts)
+
     def take_over_from(self, other_tab: 'Tab') -> None:
         self.name, self.cwd = other_tab.name, other_tab.cwd
         self.enabled_layouts = list(other_tab.enabled_layouts)
@@ -127,6 +157,7 @@ class Tab:  # {{{
         for window in self.windows:
             window.change_tab(self)
             attach_window(self.os_window_id, self.id, window.id)
+        self.active_window_changed()
         self.relayout()
 
     def _set_current_layout(self, layout_name: str) -> None:
@@ -137,7 +168,11 @@ class Tab:  # {{{
 
     def startup(self, session_tab: 'SessionTab') -> None:
         for cmd in session_tab.windows:
-            self.new_special_window(cmd)
+            if isinstance(cmd, SpecialWindowInstance):
+                self.new_special_window(cmd)
+            else:
+                from .launch import launch
+                launch(get_boss(), cmd.opts, cmd.args, target_tab=self, force_target_tab=True)
         self.windows.set_active_window_group_for(self.windows.all_windows[session_tab.active_window_idx])
 
     def serialize_state(self) -> Dict[str, Any]:
@@ -147,6 +182,9 @@ class Tab:  # {{{
             'window_list': self.windows.serialize_state(),
             'current_layout': self._current_layout_name,
             'last_used_layout': self._last_used_layout,
+            'layout_opts': self.current_layout.layout_opts,
+            'layout_state': self.current_layout.layout_state,
+            'enabled_layouts': self.enabled_layouts,
             'name': self.name,
         }
 
@@ -174,6 +212,14 @@ class Tab:  # {{{
     def title(self) -> str:
         return cast(str, getattr(self.active_window, 'title', appname))
 
+    @property
+    def number_of_windows_with_running_programs(self) -> int:
+        ans = 0
+        for window in self:
+            if window.has_running_program:
+                ans += 1
+        return ans
+
     def set_title(self, title: str) -> None:
         self.name = title or ''
         self.mark_tab_bar_dirty()
@@ -199,7 +245,7 @@ class Tab:  # {{{
             ly = self.current_layout
             self.borders(
                 all_windows=self.windows,
-                current_layout=ly, extra_blank_rects=tm.blank_rects,
+                current_layout=ly, tab_bar_rects=tm.tab_bar_rects,
                 draw_window_borders=(ly.needs_window_borders and self.windows.num_visble_groups > 1) or ly.must_draw_borders
             )
             if w is not None:
@@ -208,6 +254,7 @@ class Tab:  # {{{
     def create_layout_object(self, name: str) -> Layout:
         return create_layout_object_for(name, self.os_window_id, self.id)
 
+    @ac('lay', 'Go to the next enabled layout')
     def next_layout(self) -> None:
         if len(self.enabled_layouts) > 1:
             for i, layout_name in enumerate(self.enabled_layouts):
@@ -220,29 +267,60 @@ class Tab:  # {{{
             self._set_current_layout(nl)
             self.relayout()
 
+    @ac('lay', 'Go to the previously used layout')
     def last_used_layout(self) -> None:
         if len(self.enabled_layouts) > 1 and self._last_used_layout and self._last_used_layout != self._current_layout_name:
             self._set_current_layout(self._last_used_layout)
             self.relayout()
 
+    @ac('lay', '''
+        Switch to the named layout
+
+        For example::
+
+            map f1 goto_layout tall
+        ''')
     def goto_layout(self, layout_name: str, raise_exception: bool = False) -> None:
         layout_name = layout_name.lower()
         if layout_name not in self.enabled_layouts:
             if raise_exception:
                 raise ValueError(layout_name)
-            log_error('Unknown or disabled layout: {}'.format(layout_name))
+            log_error(f'Unknown or disabled layout: {layout_name}')
             return
         self._set_current_layout(layout_name)
         self.relayout()
 
+    @ac('lay', '''
+        Toggle the named layout
+
+        Switches to the named layout if another layout is current, otherwise
+        switches to the last used layout. Useful to "zoom" a window temporarily
+        by switching to the stack layout. For example::
+
+            map f1 toggle_layout stack
+        ''')
+    def toggle_layout(self, layout_name: str) -> None:
+        if self._current_layout_name == layout_name:
+            self.last_used_layout()
+        else:
+            self.goto_layout(layout_name)
+
     def resize_window_by(self, window_id: int, increment: float, is_horizontal: bool) -> Optional[str]:
-        increment_as_percent = self.current_layout.bias_increment_for_cell(is_horizontal) * increment
+        increment_as_percent = self.current_layout.bias_increment_for_cell(self.windows, window_id, is_horizontal) * increment
         if self.current_layout.modify_size_of_window(self.windows, window_id, increment_as_percent, is_horizontal):
             self.relayout()
             return None
         return 'Could not resize'
 
+    @ac('win', '''
+        Resize the active window by the specified amount
+
+        See :ref:`window_resizing` for details.
+        ''')
     def resize_window(self, quality: str, increment: int) -> None:
+        if quality == 'reset':
+            self.reset_window_sizes()
+            return
         if increment < 1:
             raise ValueError(increment)
         is_horizontal = quality in ('wider', 'narrower')
@@ -250,12 +328,15 @@ class Tab:  # {{{
         w = self.active_window
         if w is not None and self.resize_window_by(
                 w.id, increment, is_horizontal) is not None:
-            ring_bell()
+            if get_options().enable_audio_bell:
+                ring_bell()
 
+    @ac('win', 'Reset window sizes undoing any dynamic resizing of windows')
     def reset_window_sizes(self) -> None:
         if self.current_layout.remove_all_biases():
             self.relayout()
 
+    @ac('lay', 'Perform a layout specific action. See :doc:`layouts` for details')
     def layout_action(self, action_name: str, args: Sequence[str]) -> None:
         ret = self.current_layout.layout_action(action_name, args, self.windows)
         if ret is None:
@@ -268,16 +349,51 @@ class Tab:  # {{{
         use_shell: bool = False,
         cmd: Optional[List[str]] = None,
         stdin: Optional[bytes] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         cwd: Optional[str] = None,
         env: Optional[Dict[str, str]] = None,
         allow_remote_control: bool = False
     ) -> Child:
+        check_for_suitability = True
         if cmd is None:
             if use_shell:
-                cmd = resolved_shell(self.opts)
+                cmd = resolved_shell(get_options())
+                check_for_suitability = False
             else:
-                cmd = self.args.args or resolved_shell(self.opts)
+                if self.args.args:
+                    cmd = list(self.args.args)
+                else:
+                    cmd = resolved_shell(get_options())
+                    check_for_suitability = False
+        if check_for_suitability:
+            old_exe = cmd[0]
+            if not os.path.isabs(old_exe):
+                from .utils import which
+                actual_exe = which(old_exe)
+                old_exe = actual_exe if actual_exe else os.path.abspath(old_exe)
+            try:
+                is_executable = os.access(old_exe, os.X_OK)
+            except OSError:
+                pass
+            else:
+                try:
+                    st = os.stat(old_exe)
+                except OSError:
+                    pass
+                else:
+                    if stat.S_ISDIR(st.st_mode):
+                        cwd = old_exe
+                        cmd = resolved_shell(get_options())
+                    elif not is_executable:
+                        import shlex
+                        with suppress(OSError):
+                            with open(old_exe) as f:
+                                if f.read(2) == '#!':
+                                    line = f.read(4096).splitlines()[0]
+                                    cmd[:0] = shlex.split(line)
+                                else:
+                                    cmd[:0] = [resolved_shell(get_options())[0]]
+                                cmd[:0] = [kitty_exe(), '+hold']
         fenv: Dict[str, str] = {}
         if env:
             fenv.update(env)
@@ -285,12 +401,12 @@ class Tab:  # {{{
         pwid = platform_window_id(self.os_window_id)
         if pwid is not None:
             fenv['WINDOWID'] = str(pwid)
-        ans = Child(cmd, cwd or self.cwd, self.opts, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
+        ans = Child(cmd, cwd or self.cwd, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
         ans.fork()
         return ans
 
-    def _add_window(self, window: Window, location: Optional[str] = None, overlay_for: Optional[int] = None) -> None:
-        self.current_layout.add_window(self.windows, window, location, overlay_for)
+    def _add_window(self, window: Window, location: Optional[str] = None, overlay_for: Optional[int] = None, overlay_behind: bool = False) -> None:
+        self.current_layout.add_window(self.windows, window, location, overlay_for, put_overlay_behind=overlay_behind)
         self.mark_tab_bar_dirty()
         self.relayout()
 
@@ -300,7 +416,7 @@ class Tab:  # {{{
         cmd: Optional[List[str]] = None,
         stdin: Optional[bytes] = None,
         override_title: Optional[str] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         cwd: Optional[str] = None,
         overlay_for: Optional[int] = None,
         env: Optional[Dict[str, str]] = None,
@@ -308,17 +424,18 @@ class Tab:  # {{{
         copy_colors_from: Optional[Window] = None,
         allow_remote_control: bool = False,
         marker: Optional[str] = None,
-        watchers: Optional[Watchers] = None
+        watchers: Optional[Watchers] = None,
+        overlay_behind: bool = False
     ) -> Window:
         child = self.launch_child(
             use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env, allow_remote_control=allow_remote_control)
         window = Window(
-            self, child, self.opts, self.args, override_title=override_title,
+            self, child, self.args, override_title=override_title,
             copy_colors_from=copy_colors_from, watchers=watchers
         )
         # Must add child before laying out so that resize_pty succeeds
         get_boss().add_child(window)
-        self._add_window(window, location=location, overlay_for=overlay_for)
+        self._add_window(window, location=location, overlay_for=overlay_for, overlay_behind=overlay_behind)
         if marker:
             try:
                 window.set_marker(marker)
@@ -339,14 +456,10 @@ class Tab:  # {{{
             override_title=special_window.override_title,
             cwd_from=special_window.cwd_from, cwd=special_window.cwd, overlay_for=special_window.overlay_for,
             env=special_window.env, location=location, copy_colors_from=copy_colors_from,
-            allow_remote_control=allow_remote_control, watchers=special_window.watchers
+            allow_remote_control=allow_remote_control, watchers=special_window.watchers, overlay_behind=special_window.overlay_behind
         )
 
-    def close_window(self) -> None:
-        w = self.active_window
-        if w is not None:
-            self.remove_window(w)
-
+    @ac('win', 'Close all windows in the tab other than the currently active window')
     def close_other_windows_in_tab(self) -> None:
         if len(self.windows) > 1:
             active_window = self.active_window
@@ -354,6 +467,9 @@ class Tab:  # {{{
                 if window is not active_window:
                     self.remove_window(window)
 
+    def move_window_to_top_of_group(self, window: Window) -> bool:
+        return self.windows.move_window_to_top_of_group(window)
+
     def remove_window(self, window: Window, destroy: bool = True) -> None:
         self.windows.remove_window(window)
         if destroy:
@@ -384,23 +500,75 @@ class Tab:  # {{{
     def get_nth_window(self, n: int) -> Optional[Window]:
         if self.windows:
             return self.current_layout.nth_window(self.windows, n)
+        return None
 
+    @ac('win', '''
+        Focus the nth window if positive or the previously active windows if negative. When the number is larger
+        than the number of windows focus the last window. For example::
+
+            # focus the previously active window
+            map ctrl+p nth_window -1
+            # focus the first window
+            map ctrl+1 nth_window 0
+        ''')
     def nth_window(self, num: int = 0) -> None:
         if self.windows:
             if num < 0:
                 self.windows.make_previous_group_active(-num)
-            else:
-                self.current_layout.activate_nth_window(self.windows, num)
+            elif self.windows.num_groups:
+                self.current_layout.activate_nth_window(self.windows, min(num, self.windows.num_groups - 1))
             self.relayout_borders()
 
+    @ac('win', 'Focus the first window')
+    def first_window(self) -> None:
+        self.nth_window(0)
+
+    @ac('win', 'Focus the second window')
+    def second_window(self) -> None:
+        self.nth_window(1)
+
+    @ac('win', 'Focus the third window')
+    def third_window(self) -> None:
+        self.nth_window(2)
+
+    @ac('win', 'Focus the fourth window')
+    def fourth_window(self) -> None:
+        self.nth_window(3)
+
+    @ac('win', 'Focus the fifth window')
+    def fifth_window(self) -> None:
+        self.nth_window(4)
+
+    @ac('win', 'Focus the sixth window')
+    def sixth_window(self) -> None:
+        self.nth_window(5)
+
+    @ac('win', 'Focus the seventh window')
+    def seventh_window(self) -> None:
+        self.nth_window(6)
+
+    @ac('win', 'Focus the eighth window')
+    def eighth_window(self) -> None:
+        self.nth_window(7)
+
+    @ac('win', 'Focus the ninth window')
+    def ninth_window(self) -> None:
+        self.nth_window(8)
+
+    @ac('win', 'Focus the tenth window')
+    def tenth_window(self) -> None:
+        self.nth_window(9)
+
     def _next_window(self, delta: int = 1) -> None:
         if len(self.windows) > 1:
             self.current_layout.next_window(self.windows, delta)
             self.relayout_borders()
 
+    @ac('win', 'Focus the next window in the current tab')
     def next_window(self) -> None:
         self._next_window()
 
+    @ac('win', 'Focus the previous window in the current tab')
     def previous_window(self) -> None:
         self._next_window(-1)
 
@@ -416,18 +584,42 @@ class Tab:  # {{{
 
         if groups:
             return groups[0]
+        return None
+
+    def nth_active_window_id(self, n: int = 0) -> int:
+        if n <= 0:
+            return self.active_window.id if self.active_window else 0
+        ids = tuple(reversed(self.windows.active_window_history))
+        return ids[min(n - 1, len(ids) - 1)] if ids else 0
 
     def neighboring_group_id(self, which: EdgeLiteral) -> Optional[int]:
         neighbors = self.current_layout.neighbors(self.windows)
         candidates = neighbors.get(which)
         if candidates:
             return self.most_recent_group(candidates)
+        return None
 
+    @ac('win', '''
+        Focus the neighboring window in the current tab
+
+        For example::
+
+            map ctrl+left neighboring_window left
+            map ctrl+down neighboring_window bottom
+        ''')
     def neighboring_window(self, which: EdgeLiteral) -> None:
         neighbor = self.neighboring_group_id(which)
         if neighbor:
             self.windows.set_active_group(neighbor)
 
+    @ac('win', '''
+        Move the window in the specified direction
+
+        For example::
+
+            map ctrl+left move_window left
+            map ctrl+down move_window bottom
+        ''')
     def move_window(self, delta: Union[EdgeLiteral, int] = 1) -> None:
         if isinstance(delta, int):
             if self.current_layout.move_window(self.windows, delta):
@@ -438,22 +630,59 @@ class Tab:  # {{{
                 if self.current_layout.move_window_to_group(self.windows, neighbor):
                     self.relayout()
 
+    def swap_active_window_with(self, window_id: int) -> None:
+        group = self.windows.group_for_window(window_id)
+        if group is not None:
+            w = self.active_window
+            if w is not None and w.id != window_id:
+                if self.current_layout.move_window_to_group(self.windows, group.id):
+                    self.relayout()
+
+    @property
+    def all_window_ids_except_active_window(self) -> Set[int]:
+        all_window_ids = {w.id for w in self}
+        aw = self.active_window
+        if aw is not None:
+            all_window_ids.discard(aw.id)
+        return all_window_ids
+
+    @ac('win', '''
+        Focus a visible window by pressing the number of the window. Window numbers are displayed
+        over the windows for easy selection in this mode. See :opt:`visual_window_select_characters`.
+        ''')
+    def focus_visible_window(self) -> None:
+        def callback(tab: Optional[Tab], window: Optional[Window]) -> None:
+            if tab and window:
+                tab.set_active_window(window)
+
+        get_boss().visual_window_select_action(self, callback, 'Choose window to switch to', only_window_ids=self.all_window_ids_except_active_window)
+
+    @ac('win', 'Swap the current window with another window in the current tab, selected visually. See :opt:`visual_window_select_characters`')
+    def swap_with_window(self) -> None:
+        def callback(tab: Optional[Tab], window: Optional[Window]) -> None:
+            if tab and window:
+                tab.swap_active_window_with(window.id)
+        get_boss().visual_window_select_action(self, callback, 'Choose window to swap with', only_window_ids=self.all_window_ids_except_active_window)
+
+    @ac('win', 'Move active window to the top (make it the first window)')
     def move_window_to_top(self) -> None:
         n = self.windows.active_group_idx
         if n > 0:
             self.move_window(-n)
 
+    @ac('win', 'Move active window forward (swap it with the next window)')
     def move_window_forward(self) -> None:
         self.move_window()
 
+    @ac('win', 'Move active window backward (swap it with the previous window)')
     def move_window_backward(self) -> None:
         self.move_window(-1)
 
-    def list_windows(self, active_window: Optional[Window]) -> Generator[WindowDict, None, None]:
+    def list_windows(self, active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[WindowDict, None, None]:
         for w in self:
-            yield w.as_dict(is_focused=w is active_window)
+            yield w.as_dict(is_focused=w is active_window, is_self=w is self_window)
 
-    def matches(self, field: str, pat: Pattern) -> bool:
+    def matches(self, field: str, pat: 'Pattern[str]') -> bool:
         if field == 'id':
             return bool(pat.pattern == str(self.id))
         if field == 'title':
@@ -466,6 +695,10 @@ class Tab:  # {{{
     def __len__(self) -> int:
         return len(self.windows)
 
+    @property
+    def num_window_groups(self) -> int:
+        return self.windows.num_groups
+
     def __contains__(self, window: Window) -> bool:
         return window in self.windows
 
@@ -476,7 +709,7 @@ class Tab:  # {{{
         self.windows = WindowList(self)
 
     def __repr__(self) -> str:
-        return 'Tab(title={}, id={})'.format(self.name or self.title, hex(id(self)))
+        return f'Tab(title={self.name or self.title}, id={hex(id(self))})'
 
     def make_active(self) -> None:
         tm = self.tab_manager_ref()
@@ -487,14 +720,17 @@ class Tab:  # {{{
 
 class TabManager:  # {{{
 
-    def __init__(self, os_window_id: int, opts: Options, args: CLIOptions, startup_session: Optional[SessionType] = None):
+    def __init__(self, os_window_id: int, args: CLIOptions, wm_class: str, wm_name: str, startup_session: Optional[SessionType] = None):
         self.os_window_id = os_window_id
+        self.wm_class = wm_class
+        self.recent_mouse_events: Deque[TabMouseEvent] = deque()
+        self.wm_name = wm_name
         self.last_active_tab_id = None
-        self.opts, self.args = opts, args
-        self.tab_bar_hidden = self.opts.tab_bar_style == 'hidden'
+        self.args = args
+        self.tab_bar_hidden = get_options().tab_bar_style == 'hidden'
         self.tabs: List[Tab] = []
         self.active_tab_history: Deque[int] = deque()
-        self.tab_bar = TabBar(self.os_window_id, opts)
+        self.tab_bar = TabBar(self.os_window_id)
         self._active_tab_idx = 0
 
         if startup_session is not None:
@@ -508,6 +744,9 @@ class TabManager:  # {{{
 
     @active_tab_idx.setter
     def active_tab_idx(self, val: int) -> None:
+        new_active_tab_idx = max(0, min(val, len(self.tabs) - 1))
+        if new_active_tab_idx == self._active_tab_idx:
+            return
         try:
             old_active_tab: Optional[Tab] = self.tabs[self._active_tab_idx]
         except Exception:
@@ -515,7 +754,7 @@ class TabManager:  # {{{
         else:
             assert old_active_tab is not None
             add_active_id_to_history(self.active_tab_history, old_active_tab.id)
-        self._active_tab_idx = max(0, min(val, len(self.tabs) - 1))
+        self._active_tab_idx = new_active_tab_idx
         try:
             new_active_tab: Optional[Tab] = self.tabs[self._active_tab_idx]
         except Exception:
@@ -536,7 +775,7 @@ class TabManager:  # {{{
 
     @property
     def tab_bar_should_be_visible(self) -> bool:
-        return len(self.tabs) >= self.opts.tab_bar_min_tabs
+        return len(self.tabs) >= get_options().tab_bar_min_tabs
 
     def _add_tab(self, tab: Tab) -> None:
         visible_before = self.tab_bar_should_be_visible
@@ -599,6 +838,19 @@ class TabManager:  # {{{
         if len(self.tabs) > 1:
             self.set_active_tab_idx((self.active_tab_idx + len(self.tabs) + delta) % len(self.tabs))
 
+    def tab_at_location(self, loc: str) -> Optional[Tab]:
+        if loc == 'prev':
+            if self.active_tab_history:
+                old_active_tab_id = self.active_tab_history[-1]
+                for idx, tab in enumerate(self.tabs):
+                    if tab.id == old_active_tab_id:
+                        return tab
+        elif loc in ('left', 'right'):
+            delta = -1 if loc == 'left' else 1
+            idx = (len(self.tabs) + self.active_tab_idx + delta) % len(self.tabs)
+            return self.tabs[idx]
+        return None
+
     def goto_tab(self, tab_num: int) -> None:
         if tab_num >= len(self.tabs):
             tab_num = max(0, len(self.tabs) - 1)
@@ -614,20 +866,29 @@ class TabManager:  # {{{
                     self.set_active_tab_idx(idx)
                     break
 
+    def nth_active_tab(self, n: int = 0) -> Optional[Tab]:
+        if n <= 0:
+            return self.active_tab
+        tab_ids = tuple(reversed(self.active_tab_history))
+        return self.tab_for_id(tab_ids[min(n - 1, len(tab_ids) - 1)]) if tab_ids else None
+
     def __iter__(self) -> Iterator[Tab]:
         return iter(self.tabs)
 
     def __len__(self) -> int:
         return len(self.tabs)
 
-    def list_tabs(self, active_tab: Optional[Tab], active_window: Optional[Window]) -> Generator[TabDict, None, None]:
+    def list_tabs(self, active_tab: Optional[Tab], active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[TabDict, None, None]:
         for tab in self:
             yield {
                 'id': tab.id,
                 'is_focused': tab is active_tab,
                 'title': tab.name or tab.title,
                 'layout': str(tab.current_layout.name),
-                'windows': list(tab.list_windows(active_window)),
+                'layout_state': tab.current_layout.layout_state(),
+                'layout_opts': tab.current_layout.layout_opts.serialized(),
+                'enabled_layouts': tab.enabled_layouts,
+                'windows': list(tab.list_windows(active_window, self_window)),
                 'active_window_history': list(tab.windows.active_window_history),
             }
 
@@ -641,26 +902,37 @@ class TabManager:  # {{{
 
     @property
     def active_tab(self) -> Optional[Tab]:
-        return self.tabs[self.active_tab_idx] if self.tabs else None
+        try:
+            return self.tabs[self.active_tab_idx] if self.tabs else None
+        except Exception:
+            return None
 
     @property
     def active_window(self) -> Optional[Window]:
         t = self.active_tab
         if t is not None:
             return t.active_window
+        return None
+
+    @property
+    def number_of_windows_with_running_programs(self) -> int:
+        count = 0
+        for tab in self:
+            count += tab.number_of_windows_with_running_programs
+        return count
 
     @property
     def number_of_windows(self) -> int:
         count = 0
         for tab in self:
-            for window in tab:
-                count += 1
+            count += len(tab)
         return count
 
     def tab_for_id(self, tab_id: int) -> Optional[Tab]:
         for t in self.tabs:
             if t.id == tab_id:
                 return t
+        return None
 
     def move_tab(self, delta: int = 1) -> None:
         if len(self.tabs) > 1:
@@ -676,7 +948,7 @@ class TabManager:  # {{{
     def new_tab(
         self,
         special_window: Optional[SpecialWindowInstance] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         as_neighbor: bool = False,
         empty_tab: bool = False,
         location: str = 'last'
@@ -689,6 +961,8 @@ class TabManager:  # {{{
             location = 'after'
         if location == 'neighbor':
             location = 'after'
+        if location == 'default':
+            location = 'last'
         if len(self.tabs) > 1 and location != 'last':
             if location == 'first':
                 desired_idx = 0
@@ -704,28 +978,41 @@ class TabManager:  # {{{
         return self.tabs[idx]
 
     def remove(self, tab: Tab) -> None:
+        active_tab_before_removal = self.active_tab
         self._remove_tab(tab)
-        next_active_tab = -1
+        active_tab = self.active_tab
+        active_tab_needs_to_change = (active_tab is None and (active_tab_before_removal is None or active_tab_before_removal is tab)) or active_tab is tab
         while True:
             try:
                 self.active_tab_history.remove(tab.id)
             except ValueError:
                 break
 
-        if self.opts.tab_switch_strategy == 'previous':
-            while self.active_tab_history and next_active_tab < 0:
-                tab_id = self.active_tab_history.pop()
-                for idx, qtab in enumerate(self.tabs):
-                    if qtab.id == tab_id:
-                        next_active_tab = idx
-                        break
-        elif self.opts.tab_switch_strategy == 'left':
-            next_active_tab = max(0, self.active_tab_idx - 1)
+        if active_tab_needs_to_change:
+            next_active_tab = -1
+            if get_options().tab_switch_strategy == 'previous':
+                while self.active_tab_history and next_active_tab < 0:
+                    tab_id = self.active_tab_history.pop()
+                    for idx, qtab in enumerate(self.tabs):
+                        if qtab.id == tab_id:
+                            next_active_tab = idx
+                            break
+            elif get_options().tab_switch_strategy == 'left':
+                next_active_tab = max(0, self.active_tab_idx - 1)
+            elif get_options().tab_switch_strategy == 'right':
+                next_active_tab = min(self.active_tab_idx, len(self.tabs) - 1)
 
-        if next_active_tab < 0:
-            next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
+            if next_active_tab < 0:
+                next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
 
-        self._set_active_tab(next_active_tab)
+            self._set_active_tab(next_active_tab)
+        elif active_tab_before_removal is not None:
+            try:
+                idx = self.tabs.index(active_tab_before_removal)
+            except Exception:
+                pass
+            else:
+                self._active_tab_idx = idx
         self.mark_tab_bar_dirty()
         tab.destroy()
 
@@ -744,18 +1031,42 @@ class TabManager:  # {{{
                     has_activity_since_last_focus = True
             ans.append(TabBarData(
                 title, t is at, needs_attention,
-                len(t), t.current_layout.name or '',
-                has_activity_since_last_focus
+                len(t), t.num_window_groups, t.current_layout.name or '',
+                has_activity_since_last_focus, t.active_fg, t.active_bg,
+                t.inactive_fg, t.inactive_bg
             ))
         return ans
 
-    def activate_tab_at(self, x: int) -> None:
+    def handle_click_on_tab(self, x: int, button: int, modifiers: int, action: int) -> None:
         i = self.tab_bar.tab_at(x)
-        if i is not None:
-            self.set_active_tab_idx(i)
+        now = monotonic()
+        if i is None:
+            if button == GLFW_MOUSE_BUTTON_LEFT and action == GLFW_RELEASE and len(self.recent_mouse_events) > 2:
+                ci = get_click_interval()
+                prev, prev2 = self.recent_mouse_events[-1], self.recent_mouse_events[-2]
+                if (
+                    prev.button == button and prev2.button == button and
+                    prev.action == GLFW_PRESS and prev2.action == GLFW_RELEASE and
+                    prev.tab_idx is None and prev2.tab_idx is None and
+                    now - prev.at <= ci and now - prev2.at <= 2 * ci
+                ):  # double click
+                    self.new_tab()
+                    self.recent_mouse_events.clear()
+                    return
+        else:
+            if action == GLFW_PRESS and button == GLFW_MOUSE_BUTTON_LEFT:
+                self.set_active_tab_idx(i)
+            elif button == GLFW_MOUSE_BUTTON_MIDDLE and action == GLFW_RELEASE and self.recent_mouse_events:
+                p = self.recent_mouse_events[-1]
+                if p.button == button and p.action == GLFW_PRESS and p.tab_idx == i:
+                    tab = self.tabs[i]
+                    get_boss().close_tab(tab)
+        self.recent_mouse_events.append(TabMouseEvent(button, modifiers, action, now, i))
+        if len(self.recent_mouse_events) > 5:
+            self.recent_mouse_events.popleft()
 
     @property
-    def blank_rects(self) -> Tuple[Rect, ...]:
+    def tab_bar_rects(self) -> Tuple[Border, ...]:
         return self.tab_bar.blank_rects if self.tab_bar_should_be_visible else ()
 
     def destroy(self) -> None:
@@ -764,4 +1075,13 @@ class TabManager:  # {{{
         self.tab_bar.destroy()
         del self.tab_bar
         del self.tabs
+
+    def apply_options(self) -> None:
+        for tab in self:
+            tab.apply_options()
+        self.tab_bar_hidden = get_options().tab_bar_style == 'hidden'
+        self.tab_bar.apply_options()
+        self.update_tab_bar_data()
+        self.mark_tab_bar_dirty()
+        self.tab_bar.layout()
 # }}}
